
/******************************************************************************
* MODULE     : list.hpp
* DESCRIPTION: Linked lists
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_LIST_HPP
#define __MML_LIST_HPP
#include "base.hpp"
namespace mml {
#define TMPL_DEF template<typename C>
#define TMPL template<typename C>
TMPL class list_rep;
TMPL class list;

/******************************************************************************
* List class and its representation class
******************************************************************************/

TMPL_DEF
class list_rep REP_STRUCT {
  C       item;
  list<C> next;

public:
  list_rep<C> (C item2, list<C> next2): item(item2), next(next2) {}
  friend class list<C>;
  friend C car LESSGTR (const list<C>& l);
  friend C& car LESSGTR (list<C>& l);
  friend list<C> cdr LESSGTR (const list<C>& l);
  friend list<C>& cdr LESSGTR (list<C>& l);
};

TMPL_DEF
class list {
INDIRECT_PROTO_1 (list, list_rep, C)
private:
  list (C item, list<C> next): rep (new list_rep<C> (item, next)) {}
public:
  list (): rep (NULL) {}
  list (C c1): rep (new list_rep<C> (c1, list<C> ())) {}
  list (C c1, C c2): rep (new list_rep<C> (c1, list<C> (c2))) {}
  list (C c1, C c2, C c3): rep (new list_rep<C> (c1, list<C> (c2, c3))) {}
  list (C c1, C c2, C c3, C c4):
    rep (new list_rep<C>(c1, list<C> (c2, c3, c4))) {}
  C& operator[] (nat i) {
    assert (rep != NULL);
    return i==0? rep->item: rep->next[i-1]; }
  C  operator[] (nat i) const {
    assert (rep != NULL);
    return i==0? rep->item: rep->next[i-1]; }
  friend list<C> cons LESSGTR (const C& c, const list<C>& l);
  friend C car LESSGTR (const list<C>& l);
  friend C& car LESSGTR (list<C>& l);
  friend list<C> cdr LESSGTR (const list<C>& l);
  friend list<C>& cdr LESSGTR (list<C>& l);
  friend bool nil LESSGTR (const list<C>& l);
};
INDIRECT_IMPL_1 (list, list_rep, typename, C)

/******************************************************************************
* Basic constructors and accessors
******************************************************************************/

TMPL list<C> cons (const C& c, const list<C>& l) { return list<C> (c, l); }
TMPL C car (const list<C>& l) { return l.rep->item; }
TMPL C& car (list<C>& l) { return l.rep->item; }
TMPL list<C> cdr (const list<C>& l) { return l.rep->next; }
TMPL list<C>& cdr (list<C>& l) { return l.rep->next; }
TMPL bool nil (const list<C>& l) { return l.rep == NULL; }
TMPL int N (const list<C>& l) { return nil (l)? 0: N (cdr (l)) + 1; }
TMPL bool atom (const list<C>& l) { return N(l)==1; }

/******************************************************************************
* Printing lists
******************************************************************************/

TMPL ostream&
print_lisp (ostream& out, const list<C>& l) {
  out << "(list";
  list<C> counter= l;
  while (!nil (counter)) {
    out << " ";
    print_lisp (out, car (counter));
    counter= cdr (counter);
  }
  return out << ")";
}

TMPL ostream&
operator << (ostream& out, const list<C>& l) {
  bool first= true;
  out << "[ ";
  list<C> counter= l;
  while (!nil (counter)) {
    if (!first) out << ", ";
    print_lisp (out, car (counter));
    counter= cdr (counter);
    first= false;
  }
  if (!first) out << " ";
  return out << "]";
}

/******************************************************************************
* Operations on lists
******************************************************************************/

TMPL list<C>
copy (const list<C>& l) {
  if (nil (l)) return l;
  return cons (car (l), cdr (l));
}

TMPL list<C>
operator * (const list<C>& l1, const list<C>& l2) {
  if (nil (l1)) return copy (l2);
  return cons (car (l1), cdr (l1) * l2);
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL bool
operator == (const list<C>& l1, const list<C>& l2) {
  if (nil (l1)) return nil (l2);
  return car (l1) == car (l2) && cdr (l1) == cdr (l2);
}

TMPL bool
operator != (const list<C>& l1, const list<C>& l2) {
  if (nil (l1)) return !nil (l2);
  return car (l1) != car (l2) || cdr (l1) != cdr (l2);
}

TMPL bool
operator < (const list<C>& l1, const list<C>& l2) {
  if (nil (l1)) return !nil (l2);
  if (car (l1) < car (l2)) return true;
  return car (l1) == car (l2) && cdr (l1) < cdr (l2);
}

TMPL bool
operator <= (const list<C>& l1, const list<C>& l2) {
  if (nil (l1) || car (l1) < car (l2)) return true;
  return car (l1) == car (l2) && cdr (l1) <= cdr (l2);
}

TMPL bool
operator > (const list<C>& l1, const list<C>& l2) {
  if (nil (l1)) return !nil (l2);
  if (car (l1) > car (l2)) return true;
  return car (l1) == car (l2) && cdr (l1) > cdr (l2);
}

TMPL bool
operator >= (const list<C>& l1, const list<C>& l2) {
  if (nil (l1) || car (l1) > car (l2)) return true;
  return car (l1) == car (l2) && cdr (l1) >= cdr (l2);
}

#undef TMPL_DEF
#undef TMPL
} // namespace mml
#endif // __MML_LIST_HPP
