
/******************************************************************************
* MODULE     : table.hpp
* DESCRIPTION: Hash tables
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_TABLE_HPP
#define __MML_TABLE_HPP
#include "list.hpp"
#include "iterator.hpp"
namespace mml {
#define TMPL_DEF template<typename T, typename C>
#define TMPL template<typename T, typename C>
TMPL class table_rep;
TMPL class table;
TMPL class entries_iterator_rep;
TMPL class table_iterator_rep;

/******************************************************************************
* Table class and its representation class
******************************************************************************/

TMPL_DEF
class pair {
public:
  T key;
  C im;
  pair<T,C> () {}
  pair<T,C> (T key2, C im2): key (key2), im (im2) {}
};

TMPL_DEF
class table_rep REP_STRUCT {
  nat size;             // size of table (nr of entries)
  nat n;                // nr of keys (a power of two)
  C   init;             // default entry
  list<pair<T,C> >* a;  // the array of entries

public:
  inline table_rep<T,C>(C init2, nat n2=1):
    size(0), n(n2), init(init2), a(new list<pair<T,C> >[n]) {}
  inline ~table_rep<T,C> () { delete[] a; }
  void resize (nat n);
  void reset (const T& x);
  bool contains (const T& x);
  C    bracket_ro (const T& x) const;
  C&   bracket_rw (const T& x);
  void simplify ();

  friend class table<T,C>;
  friend class entries_iterator_rep<T,C>;
  friend class table_iterator_rep<T,C>;
  friend C&  I LESSGTR (table<T,C>& t);
  friend C   I LESSGTR (const table<T,C>& t);
  friend nat N LESSGTR (const table<T,C>& t);
  friend bool operator == LESSGTR (const table<T,C>& t1, const table<T,C>& t2);
};

TMPL_DEF
class table {
INDIRECT_PROTO_2 (table, table_rep, T, C)
  table (C init= C(), nat n=1):
    rep (new table_rep<T,C> (init, n)) {}
  C  operator [] (const T& x) const { return rep->bracket_ro (x); }
  C& operator () (const T& x) { return rep->bracket_rw (x); }
};
INDIRECT_IMPL_2 (table, table_rep, typename, T, typename, C)

/******************************************************************************
* Table entries
******************************************************************************/

TMPL ostream&
operator << (ostream& out, const pair<T,C>& p) {
  out << p.key << "->" << p.im;
  return out;
}

TMPL ostream&
print_lisp (ostream& out, const pair<T,C>& p) {
  return print_lisp ("->", p.key, p.im);
}

TMPL bool
operator == (const pair<T,C>& p1, const pair<T,C>& p2) {
  return (p1.key==p2.key) && (p1.im==p2.im);
}

TMPL bool
operator != (const pair<T,C>& p1, const pair<T,C>& p2) {
  return (p1.key!=p2.key) || (p1.im!=p2.im);
}

/******************************************************************************
* Basic access routines for tables
******************************************************************************/

TMPL C&
I (table<T,C>& t) {
  return t->init;
}

TMPL C
I (const table<T,C>& t) {
  return t->init;
}

TMPL nat
N (const table<T,C>& t) {
  return t->size;
}

TMPL void
table_rep<T,C>::resize (nat n2) {
  nat i, oldn= n;
  list<pair<T,C> >* olda= a;
  n= n2;
  a= new list<pair<T,C> >[n];
  for (i=0; i<oldn; i++) {
    list<pair<T,C> > l (olda [i]);
    while (!nil (l)) {
      list<pair<T,C> >& newl= a [hash (car(l).key) & (n-1)];
      newl= cons (car (l), newl);
      l= cdr (l);
    }
  }
  delete[] olda;
}

TMPL bool
table_rep<T,C>::contains (const T& x) {
  list<pair<T,C> >  l (a [hash (x) & (n-1)]);
  while (!nil (l)) {
    if (car (l).key==x) return true;
    l= cdr (l);
  }
  return false;
}

TMPL C&
table_rep<T,C>::bracket_rw (const T& x) {
  register nat hv= hash(x);
  list<pair<T,C> > l (a [hv & (n-1)]);
  while (!nil (l)) {
    if (car (l).key == x) return car (l).im;
    l= cdr (l);
  }
  if (size>=n) resize (n<<1);
  list<pair<T,C> >& rl= a [hv & (n-1)];
  rl= cons (pair<T,C> (x, init), rl);
  size ++;
  return car (rl).im;
}

TMPL C
table_rep<T,C>::bracket_ro (const T& x) const {
  list<pair<T,C> > l (a [hash (x) & (n-1)]);
  while (!nil (l)) {
    if (car (l).key == x) return car (l).im;
    l= cdr (l);
  }
  return init;
}

TMPL void
table_rep<T,C>::reset (const T& x) {
  list<pair<T,C> > *l= &(a [hash (x) & (n-1)]);
  while (!nil (*l)) {
    if (car (*l).key == x) {
      *l= cdr (*l);
      size --;
      if (size < (n>>1)) resize (n>>1);
      return;
    }
    l= &cdr (*l);
  }
}

TMPL void
table_rep<T,C>::simplify () {
  for (nat i=0; i<n; i++) {
    list<pair<T,C> > *l= &(a[i]);
    while (!nil (*l)) {
      if (car (*l).im == init) {
	*l= cdr (*l);
	size --;
      }
      else l= &cdr (*l);
    }
  }
  while (size < (n>>1)) resize (n>>1);
}

/******************************************************************************
* Iterators
******************************************************************************/

TMPL_DEF
class entries_iterator_rep: public iterator_rep<C> {
  table<T,C> t;        // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= cdr (l); spool (); }
  T current () { return car (l).key; }

public:
  entries_iterator_rep (table<T,C> t2): t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<T>
entries (const table<T,C>& t) {
  return iterator<C> (new entries_iterator_rep<T,C> (t), false);
}

TMPL_DEF
class table_iterator_rep: public iterator_rep<pair<T,C> > {
  table<T,C> t;        // the table to traverse
  nat i;               // current entry in table
  list<pair<T,C> > l;  // current list

protected:
  void spool () {
    while (nil (l)) {
      i++;
      if (i >= t->n) break;
      l= t->a[i];
    }
  }
  bool is_busy () { return i<t->n; }
  void advance () { l= cdr (l); spool (); }
  pair<T,C> current () { return car (l); }

public:
  table_iterator_rep (table<T,C> t2): t(t2), i(0), l(t->a[0]) { spool (); }
};

TMPL iterator<pair<T,C> >
iterate (const table<T,C>& t) {
  return iterator<pair<T,C> > (new table_iterator_rep<T,C> (t), false);
}

/******************************************************************************
* Outputting tables
******************************************************************************/

TMPL ostream&
operator << (ostream& out, const table<T,C>& t) {
  bool flag= false;
  out << "{ ";
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it) {
    if (flag) out << ", ";
    out << *it;
    flag= true;
  }
  if (flag) out << " ";
  return out << "}";
}

TMPL ostream&
print_lisp (ostream& out, const table<T,C>& t) {
  out << "(table";
  for (iterator<pair<T,C> > it= iterate (t); busy (it); ++it) {
    out << " ";
    print_lisp (*it);
  }
  return out << ")";
}

/******************************************************************************
* Vectorial operations on tables
******************************************************************************/

#define TABLE_UNARY(op)					\
TMPL table<T,C>						\
operator op (const table<T,C>& t) {			\
  table<T,C> r (I(t));					\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    r(*it)= op t[*it];					\
  r->simplify ();					\
  return r;						\
}

#define TABLE_BINARY(op)					\
TMPL table<T,C>							\
operator op (const table<T,C>& t, const table<T,C>& u) {	\
  table<T,C> r (I(t) op I(u));					\
  for (iterator<T> it= entries (t); busy (it); ++it)		\
    r(*it)= t[*it] op u[*it];					\
  for (iterator<T> it= entries (u); busy (it); ++it)		\
    if (!(t->contains (*it)))					\
      r(*it)= t[*it] op u[*it];					\
  r->simplify ();						\
  return r;							\
}

#define TABLE_SCALAR(op)				\
TMPL table<T,C>						\
operator op (const table<T,C>& t, const C& c) {		\
  table<T,C> r (I(t));					\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    r(*it)= t[*it] op c;				\
  r->simplify ();					\
  return r;						\
}

#define TABLE_INLINE_BINARY(op)				\
TMPL table<T,C>&					\
operator op (table<T,C>& t, const table<T,C>& u) {	\
  I(t) op I(u);						\
  for (iterator<T> it= entries (u); busy (it); ++it)	\
    t(*it) op u[*it];					\
  t->simplify ();					\
  return t;						\
}

#define TABLE_INLINE_SCALAR(op)				\
TMPL table<T,C>&					\
operator op (table<T,C>& t, const C& c) {		\
  I(t) op c;						\
  for (iterator<T> it= entries (t); busy (it); ++it)	\
    t(*it) op c;					\
  t->simplify ();					\
  return t;						\
}

TABLE_UNARY(-);
TABLE_BINARY(+);
TABLE_BINARY(-);
TABLE_BINARY(*);
TABLE_BINARY(/);
TABLE_BINARY(%);
TABLE_SCALAR(*);
TABLE_SCALAR(/);
TABLE_SCALAR(%);
TABLE_INLINE_BINARY(+=);
TABLE_INLINE_BINARY(-=);
TABLE_INLINE_BINARY(*=);
TABLE_INLINE_BINARY(/=);
TABLE_INLINE_BINARY(%=);
TABLE_INLINE_SCALAR(*=);
TABLE_INLINE_SCALAR(/=);
TABLE_INLINE_SCALAR(%=);

TMPL table<T,C>
operator * (const C& c, const table<T,C>& t) {
  return t * c;
}

/******************************************************************************
* Comparison routines for tables
******************************************************************************/

#define TABLE_TEST(op)							\
TMPL bool								\
operator op (const table<T,C>& t, const table<T,C>& u) {		\
  for (iterator<T> it= entries (t); busy (it); ++it)			\
    if (!(t[*it] op u[*it])) return false;				\
  for (iterator<T> it= entries (u); busy (it); ++it)			\
    if (!(t->contains (*it) || (t[*it] op u[*it]))) return false;	\
  return true;								\
}

TABLE_TEST(==);
TABLE_TEST(<=);
TABLE_TEST(<);
TABLE_TEST(>=);
TABLE_TEST(>);

TMPL bool
operator != (const table<T,C>& t1, const table<T,C>& t2) {
  return !(t1 == t2);
}

#undef TMPL_DEF
#undef TMPL
} // namespace mml
#endif // __MML_TABLE_HPP
