
/******************************************************************************
* MODULE     : polynomial.hpp
* DESCRIPTION: Fixed size polynomials
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_POLYNOMIAL_HPP
#define __MML_POLYNOMIAL_HPP
#include "jet.hpp"
namespace mml {
#define TMPL template<typename C, typename K>
TMPL class polynomial;

/******************************************************************************
* Different types of supported polynomials
******************************************************************************/

class polynomial_dense {};

/******************************************************************************
* Dense polynomials
******************************************************************************/

template<typename C, typename K=polynomial_dense>
class polynomial_rep REP_STRUCT {
public: // should be protected
  C*  a; // size of a can be larger than n due to normalization
  nat n;

public:
  void normalize () { while ((n != 0) && (a[n-1] == 0)) n--; }
  polynomial_rep (C* a2, nat n2): a (a2), n (n2) { normalize (); }
  virtual ~polynomial_rep () { delete[] a; }
  void extend (nat l) {
    C* b= new C[l];
    jet_copy (b, a, n);
    jet_clear (b+n, l-n);
    delete[] a;
    a= b;
    n= l;
  }
  friend class polynomial<C,K>;
};

template<typename C, typename K=polynomial_dense>
class polynomial {
INDIRECT_PROTO_2 (polynomial, polynomial_rep, C, K)
public:
  polynomial () {
    rep= new polynomial_rep<C,K> (new C[0], 0);
  }
  polynomial (C x, nat n=0) {
    C* a= new C[n+1];
    jet_clear (a, n);
    a[n]= x;
    rep= new polynomial_rep<C,K> (a, n+1);
  }
  polynomial (C* a, nat n) {
    rep= new polynomial_rep<C,K> (a, n);
  }
  C operator[] (nat i) const { return rep->a[i]; }
};
INDIRECT_IMPL_2 (polynomial, polynomial_rep, typename, C, typename, K)

TMPL nat   dim (const polynomial<C,K>& P) { return P->n; }
TMPL int   deg (const polynomial<C,K>& P) { return ((int) P->n) - 1; }
TMPL C*    seg (const polynomial<C,K>& P) { return P->a; }
TMPL char* var (const polynomial<C,K>& P) { return "x"; }

/******************************************************************************
* Routines on dense polynomials
******************************************************************************/

TMPL ostream&
print_lisp (ostream& out, const polynomial<C,K>& P) {
  out << "(+";
  int i, len= deg(P);
  for (i=len; i>=0; i--) {
    out << " (* ";
    (void) print_lisp (out, P[i]);
    out << " (^ " << var (P) << " " << i << "))";
  }
  out << ")";
  return out;
}

TMPL ostream&
operator << (ostream& out, const polynomial<C,K>& P) {
  bool first= true;
  int i, len= deg(P);
  if (len<0) return out << "0";
  for (i=len; i>=0; i--)
    if (P[i] != 0) {
      if (!first) out << " + ";
      if (i == 0) out << P[i];
      else {
	if (P[i] == -1) out << "-";
	else if (P[i] != 1) out << P[i] << " ";
	out << var (P);
	if (i > 1) out << "^" << i;
      }
      first= false;
    }
  return out;
}

/******************************************************************************
* Basic arithmetic operations
******************************************************************************/

TMPL polynomial<C,K>
operator + (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_add<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_copy<C> (r+m, seg(P2)+m, n-m);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>
operator - (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  nat m= min (dim (P1), dim (P2));
  nat n= max (dim (P1), dim (P2));
  C* r= new C[n];
  jet_sub<C> (r, seg (P1), seg (P2), m);
  if (dim (P1) > m) jet_copy<C> (r+m, seg(P1)+m, n-m);
  if (dim (P2) > m) jet_neg<C> (r+m, seg(P2)+m, n-m);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>
operator - (const polynomial<C,K>& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_neg<C> (r, seg (P), n);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>
operator * (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 == 0 || n2 == 0) return polynomial<C,K> ();
  C* r= new C[n1+n2-1];
  jet_mul<C> (r, seg (P1), seg (P2), n1, n2);
  return polynomial<C,K> (r, n1+n2-1);
}

TMPL polynomial<C,K>
operator * (const polynomial<C,K>& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>
operator * (const C& c, const polynomial<C,K>& P) {
  nat n= dim (P);
  C* r= new C[n];
  jet_mul<C> (r, seg (P), c, n);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>
operator / (const polynomial<C,K>& P, const C& c) {
  nat n= dim (P);
  C* r= new C[n];
  jet_div<C> (r, seg (P), c, n);
  return polynomial<C,K> (r, n);
}

TMPL polynomial<C,K>&
operator += (polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2) P1->extend (n2);
  jet_add<C> (seg (P1), seg (P2), n2);
  if (n1 == n2) P1->normalize ();
  return P1;
}

TMPL polynomial<C,K>&
operator -= (polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  nat n1= dim (P1), n2= dim (P2);
  if (n1 < n2) P1->extend (n2);
  jet_sub<C> (seg (P1), seg (P2), n2);
  if (n1 == n2) P1->normalize ();
  return P1;
}

TMPL polynomial<C,K>&
operator *= (polynomial<C,K>& P, const C& c) {
  jet_mul<C> (seg (P), c, dim (P));
  P->normalize ();
  return P;
}

TMPL polynomial<C,K>&
operator /= (polynomial<C,K>& P, const C& c) {
  jet_div<C> (seg (P), c, dim (P));
  P->normalize ();
  return P;
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL bool
operator == (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return dim (P1) == dim (P2) && jet_cmp (seg (P1), seg (P2), dim (P1)) == 0;
}

TMPL bool
operator != (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return dim (P1) != dim (P2) || jet_cmp (seg (P1), seg (P2), dim (P1)) != 0;
}

TMPL int
sign (const polynomial<C,K>& P) {
  nat n= dim (P);
  if (n == 0) return 0;
  return sign (P[n-1]);
}

TMPL bool
operator < (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return compare (P1, P2) < 0;
}

TMPL bool
operator <= (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return compare (P1, P2) <= 0;
}

TMPL bool
operator > (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return compare (P1, P2) > 0;
}

TMPL bool
operator >= (const polynomial<C,K>& P1, const polynomial<C,K>& P2) {
  return compare (P1, P2) >= 0;
}

#undef TMPL
} // namespace mml
#endif // __MML_POLYNOMIAL_HPP
