
/******************************************************************************
* MODULE     : implicit_series.hpp
* DESCRIPTION: Implicit series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_IMPLICIT_SERIES_HPP
#define __MML_IMPLICIT_SERIES_HPP
#include "polynomial.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename K=series_dense>
#define TMPL template<typename C, typename K>

/******************************************************************************
* Implicit series
******************************************************************************/

TMPL_DEF
class implicit_series_rep: public series_rep<C,K> {
  series<C,K> eq;
public:
  implicit_series_rep () {}
  C next () { return eq[n]; }
  ostream& print_lisp (ostream& out) { return out << "implicit_series"; }
  C& initial (nat n2) {
    if (n2>=n) { n=n2+1; set_precision (n); }
    return a[n2]; }
  series<C,K>& equation () { return eq; }
};

TMPL series<C,K>
implicit_series () {
  return series<C,K> (new implicit_series_rep<C,K> (), false);
}

TMPL_DEF
class build_series_rep: public series_rep<C,K> {
  series<C,K> f;
public:
  build_series_rep (series<C,K> f2, series<C,K> eq):
    f (f2) { f->equation ()= eq; }
  ~build_series_rep () { f->equation ()= 0; }
  C next () { return f[n]; }
  ostream& print_lisp (ostream& out) { return out << "implicit_series"; }
};

TMPL series<C,K>
build (const series<C,K>& f, const series<C,K>& eq) {
  return series<C,K> (new build_series_rep<C,K> (f, eq), false);
}

TMPL_DEF
class system_series_rep: public series_rep<C,K> {
  series<C,K>  f;
  series<C,K>* comp;
  nat          dim;
public:
  system_series_rep (series<C,K> f2, series<C,K>* c, series<C,K>* eq, nat d):
    f (f2), comp (c), dim (d)
  {
    for (nat i=0; i<dim; i++)
      comp[i]->equation ()= eq[i];
    delete[] eq;
  }
  ~system_series_rep () {
    for (nat i=0; i<dim; i++)
      comp[i]->equation ()= 0;
    delete[] comp;
  }
  C next () { return f[n]; }
  ostream& print_lisp (ostream& out) { return out << "implicit_series"; }
};

TMPL series<C,K>
build (const series<C,K>& f, series<C,K>* comp, series<C,K>* eq, nat dim) {
  return series<C,K> (new system_series_rep<C,K> (f, comp, eq, dim), false);
}

/******************************************************************************
* Division, square root, exponentiation and logarithm
******************************************************************************/

TMPL series<C,K>
operator / (const series<C,K>& f, const series<C,K>& g) {
  if (g[0] == 0) {
    nat val= 0;
    for (val=0; val<series_zerotest_order && g[val]==0; val++)
      assert (f[val] == 0);
    assert (val<series_zerotest_order);
    return (f>>val) / (g>>val);
  }
  series<C,K> h= implicit_series<C,K> ();
  return build (h, (f - ((h * (g>>1)) << 1)) / g[0]);
}

TMPL series<C,K>
sqrt (const series<C,K>& f) {
  if (f[0] == 0) {
    nat val= 0;
    for (val=0; val<series_zerotest_order && f[val]==0; val++);
    assert (val<series_zerotest_order && (val&1) == 0);
    return sqrt (f >> val) << (val/2);
  }
  series<C,K> g= implicit_series<C,K> ();
  g(0)= sqrt (f[0]);
  return build (g, integrate (derive (f) / (2*g)));
}

TMPL series<C,K>
pow (const series<C,K>& f, const C& c) {
  series<C,K> g= implicit_series<C,K> ();
  g(0)= pow (f[0], c);
  return build (g, integrate (c * derive (f) * g / f));
}

TMPL series<C,K>
exp (const series<C,K>& f) {
  series<C,K> g= implicit_series<C,K> ();
  g(0)= exp (f[0]);
  return build (g, integrate (g * derive (f)));
}

TMPL series<C,K>
log (const series<C,K>& f) {
  series<C,K> g= implicit_series<C,K> ();
  g(0)= log (f[0]);
  return build (g, integrate (derive (f) / f));
}

/******************************************************************************
* Trigonometric functions
******************************************************************************/

TMPL series<C,K>
sin (const series<C,K>& f) {
  series<C,K> df= derive (f);
  series<C,K>* sys= new series<C,K>[2];
  series<C,K>* eq = new series<C,K>[2];
  sys[0]= implicit_series<C,K> ();
  sys[1]= implicit_series<C,K> ();
  sys[0](0)= sin (f[0]);
  sys[1](0)= cos (f[0]);
  eq[0]= integrate (df * sys[1]);
  eq[1]= integrate (-df * sys[0]);
  return build (sys[0], sys, eq, 2);
}

TMPL series<C,K>
cos (const series<C,K>& f) {
  series<C,K> df= derive (f);
  series<C,K>* sys= new series<C,K>[2];
  series<C,K>* eq = new series<C,K>[2];
  sys[0]= implicit_series<C,K> ();
  sys[1]= implicit_series<C,K> ();
  sys[0](0)= sin (f[0]);
  sys[1](0)= cos (f[0]);
  eq[0]= integrate (df * sys[1]);
  eq[1]= integrate (-df * sys[0]);
  return build (sys[1], sys, eq, 2);
}

TMPL series<C,K>
tan (const series<C,K>& f) {
  series<C,K> df= derive (f);
  series<C,K>* sys= new series<C,K>[2];
  series<C,K>* eq = new series<C,K>[2];
  sys[0]= implicit_series<C,K> ();
  sys[1]= implicit_series<C,K> ();
  sys[0](0)= sin (f[0]);
  sys[1](0)= cos (f[0]);
  eq[0]= integrate (df * sys[1]);
  eq[1]= integrate (-df * sys[0]);
  return build (sys[0]/sys[1], sys, eq, 2);
}

#undef TMPL_DEF
#undef TMPL
} // namespace mml
#endif // __MML_IMPLICIT_SERIES_HPP
