
/******************************************************************************
* MODULE     : series.hpp
* DESCRIPTION: Relaxed series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_SERIES_HPP
#define __MML_SERIES_HPP
#include "polynomial.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename K=series_dense>
#define TMPL template<typename C, typename K>
TMPL class series;

const nat series_printing_order= 10;
const nat series_zerotest_order= 25;

/******************************************************************************
* Different types of supported series
******************************************************************************/

class series_dense {};

/******************************************************************************
* Dense series
******************************************************************************/

TMPL_DEF
class series_rep REP_STRUCT {
protected:
  C*  a;  // coefficients
  nat n;  // number of computed coefficients
  nat l;  // number of allocated coefficients
          // a[n],...,a[l-1] may be used by relaxed computations

  inline series_rep (): a (new C[0]), n (0), l (0) {}
  inline virtual ~series_rep () { delete[] a; }
  virtual C next () = 0;

public:
  virtual ostream& print_lisp (ostream& out) = 0;
  virtual void set_precision (nat l2);
  virtual void increase_precision (nat l2=0);
  virtual C& initial (nat n);
  virtual series<C,K>& equation ();
  friend class series<C,K>;
};

TMPL_DEF
class series {
INDIRECT_PROTO_2 (series, series_rep, C, K)
public:
  series (int c= 0, nat deg= 0); // FIXME: how to avoid definition if C==int
  series (C c, nat deg= 0);
  series (polynomial<C> P);
  series (C (*coeffs) (nat), char* name);
  C  operator [] (nat n) const;
  C& operator () (nat n) const;
  C* operator () (nat start, nat end) const;
};
INDIRECT_IMPL_2 (series, series_rep, typename, C, typename, K)

TMPL char* var (const series<C,K>& P) { return "z"; }

/******************************************************************************
* Routines on dense series
******************************************************************************/

TMPL void
series_rep<C,K>::set_precision (nat l2) {
  if (l2 <= l) return;
  C* b= new C[l2];
  jet_copy (b, a, l);
  jet_clear (b+l, l2-l);
  delete[] a;
  a= b;
  l= l2;
}

TMPL void
series_rep<C,K>::increase_precision (nat l2) {
  l2= max (l2, l);
  set_precision (max ((nat) 1, l2 << 1));
}

TMPL C&
series_rep<C,K>::initial (nat i) {
  fatal_error ("Not an implicit series",
	       "series_rep<C,K>::initial",
	       "series.hpp");
  return a[0];
}

TMPL series<C,K>&
series_rep<C,K>::equation () {
  fatal_error ("Not an implicit series",
	       "series_rep<C,K>::initial",
	       "series.hpp");
  return *((series<C,K>*) ((void*) this));
}

TMPL C
series<C,K>::operator [] (nat n) const {
  if (n <  rep->n) return rep->a[n];
  if (n >= rep->l) rep->increase_precision (n);
  while (rep->n <= n) {
    rep->a[rep->n]= rep->next ();
    rep->n++;
  }
  // cout << (*this) << " [" << n << "] -> " << rep->a[n] << "\n";
  return rep->a[n];
}

TMPL C*
series<C,K>::operator () (nat start, nat end) const {
  if (end <= rep->n) return rep->a + start;
  if (end >= rep->l) rep->increase_precision (end);
  while (rep->n < end) {
    rep->a[rep->n]= rep->next ();
    rep->n++;
  }
  return rep->a + start;
}

TMPL C&
series<C,K>::operator () (nat n) const {
  return rep->initial (n);
}

TMPL ostream&
print_lisp (ostream& out, const series<C,K>& f) {
  return f->print_lisp (out);
}

TMPL ostream&
operator << (ostream& out, const series<C,K>& f) {
  nat i;
  bool first= true;
  for (i=0; i<series_printing_order; i++)
    if (f[i] != 0) {
      if (!first) out << " + ";
      if (i == 0) out << f[i];
      else {
	if (f[i] == -1) out << "-";
	else if (f[i] != 1) out << f[i] << " ";
	out << var (f);
	if (i > 1) out << "^" << i;
      }
      first= false;
    }
  if (!first) out << " + ";
  return out << "O(" << var (f) << "^" << series_printing_order << ")";
}

/******************************************************************************
* Explicit series
******************************************************************************/

TMPL_DEF
class polynomial_series_rep: public series_rep<C,K> {
  polynomial<C> P;
public:
  polynomial_series_rep (polynomial<C> P2): P (P2) {}
  C next () { return n < dim(P)? P[n]: 0; }
  ostream& print_lisp (ostream& out) { return mml::print_lisp (out, P); }
};

TMPL
series<C,K>::series (int c, nat deg) {
  rep= new polynomial_series_rep<C,K> (polynomial<C> (C(c), deg));
}

TMPL
series<C,K>::series (C c, nat deg) {
  rep= new polynomial_series_rep<C,K> (polynomial<C> (c, deg));
}

TMPL
series<C,K>::series (polynomial<C> P) {
  rep= new polynomial_series_rep<C,K> (P);
}

TMPL_DEF
class func_series_rep : public series_rep<C,K> {
  C (*coeffs) (nat);
  char* name;
public:
  func_series_rep (C (*coeffs2) (nat), char* name2):
    coeffs (coeffs2), name (name2) {}
  C next () { return coeffs (n); }
  ostream& print_lisp (ostream& out) { return out << name; }
};

TMPL
series<C,K>::series (C (*coeffs) (nat), char* name) {
  rep= new func_series_rep<C> (coeffs, name);
}

/******************************************************************************
* Sums and products
******************************************************************************/

TMPL_DEF
class sum_series_rep : public series_rep<C,K> {
  series<C,K> f, g;
public:
  sum_series_rep (series<C,K> f2, series<C,K> g2):
    f(f2), g(g2) {}
  C next () { return f[n] + g[n]; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "+", f, g); }
};

TMPL series<C,K>
operator + (const series<C,K>& f, const series<C,K>& g) {
  return series<C,K> (new sum_series_rep<C,K> (f, g), false);
}

TMPL_DEF
class diff_series_rep : public series_rep<C,K> {
  series<C,K> f, g;
public:
  diff_series_rep (series<C,K> f2, series<C,K> g2):
    f(f2), g(g2) {}
  C next () { return f[n] - g[n]; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "-", f, g); }
};

TMPL series<C,K>
operator - (const series<C,K>& f) {
  return series<C,K> (new diff_series_rep<C,K> (0, f), false);
}

TMPL series<C,K>
operator - (const series<C,K>& f, const series<C,K>& g) {
  return series<C,K> (new diff_series_rep<C,K> (f, g), false);
}
TMPL series<C,K> operator += (series<C,K>& f, const series<C,K>& g) {
  return f= f + g; }
TMPL series<C,K> operator -= (series<C,K>& f, const series<C,K>& g) {
  return f= f - g; }

/******************************************************************************
* Scalar multiplication and division
******************************************************************************/

TMPL_DEF
class scprod_series_rep : public series_rep<C,K> {
  series<C,K> f;
  C c;
public:
  scprod_series_rep (series<C,K> f2, C c2): f(f2), c(c2) {}
  C next () { return f[n] * c; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "*", c, f); }
};

TMPL series<C,K>
operator * (const series<C,K>& f, const C& c) {
  return series<C,K> (new scprod_series_rep<C,K> (f, c), false);
}

TMPL series<C,K>
operator * (const C& c, const series<C,K>& f) {
  return series<C,K> (new scprod_series_rep<C,K> (f, c), false);
}

TMPL_DEF
class scdiv_series_rep : public series_rep<C,K> {
  series<C,K> f;
  C c;
public:
  scdiv_series_rep (series<C,K> f2, C c2): f(f2), c(c2) {}
  C next () { return f[n] / c; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "/", c, f); }
};

TMPL series<C,K>
operator / (const series<C,K>& f, const C& c) {
  return series<C,K> (new scdiv_series_rep<C,K> (f, c), false);
}

/******************************************************************************
* Commodity routines
* FIXME: how to avoid definitions if C==int
******************************************************************************/

TMPL series<C,K>
operator * (const series<C,K>& f, int c) {
  return series<C,K> (new scprod_series_rep<C,K> (f, C(c)), false);
}

TMPL series<C,K>
operator * (int c, const series<C,K>& f) {
  return series<C,K> (new scprod_series_rep<C,K> (f, C(c)), false);
}

TMPL series<C,K>
operator / (const series<C,K>& f, int c) {
  return series<C,K> (new scdiv_series_rep<C,K> (f, C(c)), false);
}

TMPL series<C,K>& operator *= (series<C,K>& f, const int& c) {
  return f= f * C(c); }
TMPL series<C,K>& operator /= (series<C,K>& f, const int& c) {
  return f= f / C(c); }

/******************************************************************************
* Multiplication
******************************************************************************/

TMPL_DEF
class prod_series_rep : public series_rep<C,K> {
  series<C,K> f, g;
public:
  prod_series_rep (series<C,K> f2, series<C,K> g2):
    f(f2), g(g2) {}
  C next () {
    nat i;
    C acc= f[0] * g[n];
    for (i=1; i<=n; i++)
      acc += f[i] * g[n-i];
    return acc;
  }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "*", f, g); }
};

TMPL series<C,K>
operator * (const series<C,K>& f, const series<C,K>& g) {
  return series<C,K> (new prod_series_rep<C,K> (f, g), false);
}

/******************************************************************************
* Other linear time operations
******************************************************************************/

TMPL_DEF
class shift_series_rep : public series_rep<C,K> {
  series<C,K> f;
  int shift;
public:
  shift_series_rep (series<C,K> f2, int shift2):
    f (f2), shift (shift2) {}
  C next () { return shift > ((int) n)? 0: f[n-shift]; }
  ostream& print_lisp (ostream& out) {
    if (shift>0) return mml::print_lisp (out, "<<", f, shift);
    else return mml::print_lisp (out, ">>", f, -shift); }
};

TMPL series<C,K>
operator << (const series<C,K>& f, int shift) {
  return series<C,K> (new shift_series_rep<C,K> (f, shift), false);
}

TMPL series<C,K>
operator >> (const series<C,K>& f, int shift) {
  return series<C,K> (new shift_series_rep<C,K> (f, -shift), false);
}

TMPL_DEF
class derive_series_rep : public series_rep<C,K> {
  series<C,K> f;
public:
  derive_series_rep (series<C,K> f2): f (f2) {}
  C next () { return (n+1) * f[n+1]; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "derive", f); }
};

TMPL series<C,K>
derive (const series<C,K>& f) {
  return series<C,K> (new derive_series_rep<C,K> (f), false);
}

TMPL_DEF
class integrate_series_rep : public series_rep<C,K> {
  series<C,K> f;
public:
  integrate_series_rep (series<C,K> f2): f (f2) {}
  C next () { return n==0? 0: f[n-1]/n; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "integrate", f); }
};

TMPL series<C,K>
integrate (const series<C,K>& f) {
  return series<C,K> (new integrate_series_rep<C,K> (f), false);
}

TMPL_DEF
class dilated_series_rep : public series_rep<C,K> {
  series<C,K> f;
  nat p;
public:
  dilated_series_rep (series<C,K> f2, int p2): f (f2), p (p2) {}
  C next () { return n%p == 0? f[n/p]: 0; }
  ostream& print_lisp (ostream& out) {
    return mml::print_lisp (out, "dilate", f, p); }
};

TMPL series<C,K>
dilate (const series<C,K>& f, nat p) {
  return series<C,K> (new dilated_series_rep<C,K> (f, p), false);
}

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL int
sign (const series<C,K>& f) {
  return jet_sign (f (0, series_zerotest_order), series_zerotest_order);
}

TMPL bool operator == (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) == 0; }
TMPL bool operator != (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) != 0; }
TMPL bool operator < (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) < 0; }
TMPL bool operator <= (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) <= 0; }
TMPL bool operator > (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) > 0; }
TMPL bool operator >= (const series<C,K>& f, const series<C,K>& g) {
  return compare (f, g) >= 0; }

#undef TMPL_DEF
#undef TMPL
} // namespace mml
#include "implicit_series.hpp"
#endif // __MML_SERIES_HPP
