
/******************************************************************************
* MODULE     : vector.hpp
* DESCRIPTION: Fixed size vectors
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_VECTOR_HPP
#define __MML_VECTOR_HPP
#include "base.hpp"
#include "segment.hpp"
namespace mml {
#define TMPL template<typename C, typename K>
TMPL class vector;

/******************************************************************************
* Different types of supported vectors
******************************************************************************/

class vector_dense {
public:
  static const nat def_len= 0;
};

template<nat n>
class vector_fixed {
public:
  static const nat def_len= n;
};

/*
class vector_sparse {
  static const nat def_len= 0;
};
*/

/******************************************************************************
* Dense vectors
******************************************************************************/

template<typename C, typename K=vector_dense>
class vector_rep REP_STRUCT {
public: // should be protected
  C*  a;
  nat n;
public:
  vector_rep (C* a2, nat n2): a (a2), n (n2) {}
  virtual ~vector_rep () { delete[] a; }
  friend class vector<C,K>;
};

template<typename C, typename K=vector_dense>
class vector {
INDIRECT_PROTO_2 (vector, vector_rep, C, K)
public:
  vector () {
    nat n= K::def_len;
    rep= new vector_rep<C,K> (new C[n], n);
  }
  vector (C x, nat n=K::def_len) {
    nat i;
    rep= new vector_rep<C,K> (new C[n], n);
    for (i=0; i<n; i++)
      rep->a[i]= x;
  }
  vector (C* a, nat n=K::def_len) {
    rep= new vector_rep<C,K> (a, n);
  }
  C& operator[] (nat i) { return rep->a[i]; }
  C operator[] (nat i) const { return rep->a[i]; }
};
INDIRECT_IMPL_2 (vector, vector_rep, typename, C, typename, K)

TMPL nat dim (const vector<C,K>& v) { return v->n; }
TMPL C* seg (const vector<C,K>& v) { return v->a; }

/******************************************************************************
* Fixed size vectors
******************************************************************************/

template<typename C, nat N>
class vector_rep<C,vector_fixed<N> > REP_STRUCT {
public: // should be protected
  C* a;
public:
  vector_rep (C* a2, nat dummy): a (a2) { (void) dummy; }
  virtual ~vector_rep () { delete[] a; }
  friend class vector<C,vector_fixed<N> >;
};

template<typename C, nat N> nat
dim (const vector<C,vector_fixed<N> >& v) { return N; }

/******************************************************************************
* Routines on dense vectors
******************************************************************************/

TMPL ostream&
print_lisp (ostream& out, const vector<C,K>& v) {
  int i, n= v.size();
  out << "([] ";
  for (i=0; i<n; i++) {
    out << " ";
    (void) print_lisp (out, v[i]);
  }
  out << ")";
  return out;
}

TMPL ostream&
operator << (ostream& out, const vector<C,K>& v) {
  nat i, n= dim(v);
  out << "[ ";
  for (i=0; i<n; i++) {
    if (i != 0) out << ", ";
    out << v[i];
  }
  return out << "]";
}

#define VECTOR_UNARY(name,op)			\
TMPL vector<C,K>				\
operator op (const vector<C,K>& x) {		\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), n);			\
  return vector<C,K> (r, n);			\
}

#define VECTOR_BINARY(name,op)					\
TMPL vector<C,K>						\
operator op (const vector<C,K>& x, const vector<C,K>& y) {	\
  nat n= dim (x);						\
  assert (dim (y) == n);					\
  C* r= new C[n];						\
  name<C> (r, seg (x), seg (y), n);				\
  return vector<C,K> (r, n);					\
}

#define VECTOR_SCALAR(name,op)			\
TMPL vector<C,K>				\
operator op (const vector<C,K>& x, const C c) {	\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), c, n);			\
  return vector<C,K> (r, n);			\
}

#define VECTOR_VAR_SCALAR(name,op)		\
TMPL vector<C,K>				\
operator op (const C c, const vector<C,K>& x) {	\
  nat n= dim (x);				\
  C* r= new C[n];				\
  name<C> (r, seg (x), c, n);			\
  return vector<C,K> (r, n);			\
}

#define VECTOR_INLINE_BINARY(name,op)			\
TMPL vector<C,K>&					\
operator op (vector<C,K>& x, const vector<C,K>& y) {	\
  nat n= dim (x);					\
  assert (dim (y) == n);				\
  name<C> (seg (x), seg (y), n);			\
  return x;						\
}  

#define VECTOR_INLINE_SCALAR(name,op)		\
TMPL vector<C,K>&				\
operator op (vector<C,K>& x, const C& c) {	\
  nat n= dim (x);				\
  name<C> (seg (x), c, n);			\
  return x;					\
}  

#define VECTOR_TEST(name,op)					\
TMPL bool							\
operator op (const vector<C,K>& x, const vector<C,K>& y) {	\
  nat n= dim (x);						\
  assert (dim (y) == n);					\
  return name<C> (seg (x), seg (y), n);				\
}

VECTOR_UNARY(segment_neg,-);
VECTOR_BINARY(segment_add,+);
VECTOR_BINARY(segment_sub,-);
VECTOR_BINARY(segment_mul,*);
VECTOR_BINARY(segment_div,/);
VECTOR_SCALAR(segment_mul,*);
VECTOR_VAR_SCALAR(segment_mul,*);
VECTOR_SCALAR(segment_div,/);
VECTOR_INLINE_BINARY(segment_add,+=);
VECTOR_INLINE_BINARY(segment_sub,-=);
VECTOR_INLINE_BINARY(segment_mul,*=);
VECTOR_INLINE_BINARY(segment_div,/=);
VECTOR_INLINE_SCALAR(segment_mul,*=);
VECTOR_INLINE_SCALAR(segment_div,/=);
VECTOR_TEST(segment_eq,==);
VECTOR_TEST(segment_less,<);
VECTOR_TEST(segment_lesseq,<=);
VECTOR_TEST(segment_gtr,>);
VECTOR_TEST(segment_gtreq,>=);

TMPL bool
operator != (const vector<C,K>& x, const vector<C,K>& y) {
  return !(x == y);
}

#undef TMPL
} // namespace mml
#endif // __MML_VECTOR_HPP
